@@ -32,7 +32,7 @@ use Carp::Heavy;
 use Carp;
 use Exporter;
 
-$Config::General::VERSION = 2.32;
+$Config::General::VERSION = 2.37;
 
 use vars  qw(@ISA @EXPORT_OK);
 use base qw(Exporter);
@@ -48,161 +48,45 @@ sub new {
   # define default options
   my $self = {
 	      SlashIsDirectory      => 0,
-
 	      AllowMultiOptions     => 1,
-
 	      MergeDuplicateOptions => 0,
 	      MergeDuplicateBlocks  => 0,
-
 	      LowerCaseNames        => 0,
-
+	      ApacheCompatible      => 0,
 	      UseApacheInclude      => 0,
 	      IncludeRelative       => 0,
 	      IncludeDirectories    => 0,
 	      IncludeGlob           => 0,
-
+              IncludeAgain          => 0,
 	      AutoLaunder           => 0,
-
 	      AutoTrue              => 0,
-
 	      AutoTrueFlags         => {
 					true  => '^(on|yes|true|1)$',
 					false => '^(off|no|false|0)$',
 				       },
-
 	      DefaultConfig         => {},
-
 	      level                 => 1,
-
 	      InterPolateVars       => 0,
-
 	      InterPolateEnv        => 0,
-
 	      ExtendedAccess        => 0,
-
 	      SplitPolicy           => 'guess', # also possible: whitespace, equalsign and custom
-
 	      SplitDelimiter        => 0,       # must be set by the user if SplitPolicy is 'custom'
-
 	      StoreDelimiter        => 0,       # will be set by me unless user uses 'custom' policy
-
 	      CComments             => 1,       # by default turned on
-
 	      BackslashEscape       => 0,       # by default turned off, allows escaping anything using  the backslash
-
 	      StrictObjects         => 1,       # be strict on non-existent keys in OOP mode
-
 	      StrictVars            => 1,       # be strict on undefined variables in Interpolate mode
-
 	      Tie                   => q(),      # could be set to a perl module for tie'ing new hashes
-
 	      parsed                => 0,       # internal state stuff for variable interpolation
-	      upperkey              => q(),
-	      upperkeys             => [],
-	      lastkey               => q(),
-	      prevkey               => q( ),
 	      files                 => {},      # which files we have read, if any
 	     };
 
   # create the class instance
   bless $self, $class;
 
-
   if ($#param >= 1) {
     # use of the new hash interface!
-    my %conf = @param;
-
-    # save the parameter list for ::Extended's new() calls
-    $self->{Params} = \%conf;
-
-    # be backwards compatible
-    if (exists $conf{-file}) {
-      $self->{ConfigFile} = delete $conf{-file};
-    }
-    if (exists $conf{-hash}) {
-      $self->{ConfigHash} = delete $conf{-hash};
-    }
-
-    # store input, file, handle, or array
-    if (exists $conf{-ConfigFile}) {
-      $self->{ConfigFile} = delete $conf{-ConfigFile};
-    }
-    if (exists $conf{-ConfigHash}) {
-      $self->{ConfigHash} = delete $conf{-ConfigHash};
-    }
-
-    # store search path for relative configs, if any
-    if (exists $conf{-ConfigPath}) {
-      my $configpath = delete $conf{-ConfigPath};
-      $self->{ConfigPath} = ref $configpath eq 'ARRAY' ? $configpath : [$configpath];
-    }
-
-    # handle options which contains values we are needing (strings, hashrefs or the like)
-    if (exists $conf{-String} ) {
-      if (ref(\$conf{-String}) eq 'SCALAR') {
-	if ( $conf{-String}) {
-	  $self->{StringContent} = $conf{-String};
-	}
-	delete $conf{-String};
-      }
-      else {
-	croak "Parameter -String must be a SCALAR!\n";
-      }
-    }
-
-    if (exists $conf{-Tie}) {
-      if ($conf{-Tie}) {
-	$self->{Tie} = delete $conf{-Tie};
-	$self->{DefaultConfig} = $self->_hashref();
-      }
-    }
-
-    if (exists $conf{-FlagBits}) {
-      if ($conf{-FlagBits} && ref($conf{-FlagBits}) eq 'HASH') {
-	$self->{FlagBits} = 1;
-	$self->{FlagBitsFlags} = $conf{-FlagBits};
-      }
-      delete $conf{-FlagBits};
-    }
-
-    if (exists $conf{-DefaultConfig}) {
-      if ($conf{-DefaultConfig} && ref($conf{-DefaultConfig}) eq 'HASH') {
-	$self->{DefaultConfig} = $conf{-DefaultConfig};
-      }
-      elsif ($conf{-DefaultConfig} && ref($conf{-DefaultConfig}) eq q()) {
-	$self->_read($conf{-DefaultConfig}, 'SCALAR');
-	$self->{DefaultConfig} = $self->_parse($self->_hashref(), $self->{content});
-	$self->{content} = ();
-      }
-      delete $conf{-DefaultConfig};
-    }
-
-    # handle options which may either be true or false
-    # allowing "human" logic about what is true and what is not
-    foreach my $entry (keys %conf) {
-      my $key = $entry;
-      $key =~ s/^\-//;
-      if (! exists $self->{$key}) {
-	croak "Unknown parameter: $entry => \"$conf{$entry}\" (key: <$key>)\n";
-      }
-      if ($conf{$entry} =~ /$self->{AutoTrueFlags}->{true}/io) {
-	$self->{$key} = 1;
-      }
-      elsif ($conf{$entry} =~ /$self->{AutoTrueFlags}->{false}/io) {
-	$self->{$key} = 0;
-      }
-      else {
-	# keep it untouched
-	$self->{$key} = $conf{$entry};
-      }
-    }
-
-    if ($self->{MergeDuplicateOptions}) {
-      # override if not set by user
-      if (! exists $conf{-AllowMultiOptions}) {
-	$self->{AllowMultiOptions} = 0;
-      }
-    }
+    $self->_prepare(@param);
   }
   elsif ($#param == 0) {
     # use of the old style
@@ -217,7 +101,130 @@ sub new {
     $self->{parsed} = 1;
   }
 
-  # prepare the split delimiter if needed
+  # find split policy to use for option/value separation
+  $self->_splitpolicy();
+
+  # bless into variable interpolation module if neccessary
+  $self->_blessvars();
+
+  # process as usual
+  if (!$self->{parsed}) {
+    $self->_process();
+  }
+
+  if ($self->{InterPolateVars}) {
+    $self->{config} = $self->_clean_stack($self->{config});
+  }
+
+  # bless into OOP namespace if required
+  $self->_blessoop();
+
+  return $self;
+}
+
+
+
+sub _process {
+  #
+  # call _read() and _parse() on the given config
+  my($self) = @_;
+
+  if ($self->{DefaultConfig} && $self->{InterPolateVars}) {
+    $self->{DefaultConfig} = $self->_interpolate_hash($self->{DefaultConfig}); # FIXME: _hashref() ?
+  }
+  if (exists $self->{StringContent}) {
+    # consider the supplied string as config file
+    $self->_read($self->{StringContent}, 'SCALAR');
+    $self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
+  }
+  elsif (exists $self->{ConfigHash}) {
+    if (ref($self->{ConfigHash}) eq 'HASH') {
+      # initialize with given hash
+      $self->{config} = $self->{ConfigHash};
+      $self->{parsed} = 1;
+    }
+    else {
+      croak "Config::General: Parameter -ConfigHash must be a hash reference!\n";
+    }
+  }
+  elsif (ref($self->{ConfigFile}) eq 'GLOB' || ref($self->{ConfigFile}) eq 'FileHandle') {
+    # use the file the glob points to
+    $self->_read($self->{ConfigFile});
+    $self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
+  }
+  else {
+    if ($self->{ConfigFile}) {
+      # open the file and read the contents in
+      $self->{configfile} = $self->{ConfigFile};
+      if ( file_name_is_absolute($self->{ConfigFile}) ) {
+	# look if is is an absolute path and save the basename if it is absolute
+	my ($volume, $path, undef) = splitpath($self->{ConfigFile});
+	$path =~ s#/$##; # remove eventually existing trailing slash
+	if (! $self->{ConfigPath}) {
+	  $self->{ConfigPath} = [];
+	}
+	unshift @{$self->{ConfigPath}}, catpath($volume, $path, q());
+      }
+      $self->_open($self->{configfile});
+      # now, we parse immdediately, getall simply returns the whole hash
+      $self->{config} = $self->_hashref();
+      $self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
+    }
+    else {
+      # hm, no valid config file given, so try it as an empty object
+      $self->{config} = $self->_hashref();
+      $self->{parsed} = 1;
+    }
+  }
+}
+
+
+sub _blessoop {
+  #
+  # bless into ::Extended if neccessary
+  my($self) = @_;
+  if ($self->{ExtendedAccess}) {
+    # we are blessing here again, to get into the ::Extended namespace
+    # for inheriting the methods available overthere, which we doesn't have.
+    bless $self, 'Config::General::Extended';
+    eval {
+      require Config::General::Extended;
+    };
+    if ($EVAL_ERROR) {
+      croak "Config::General: " . $EVAL_ERROR;
+    }
+  }
+#  return $self;
+}
+
+sub _blessvars {
+  #
+  # bless into ::Interpolated if neccessary
+  my($self) = @_;
+  if ($self->{InterPolateVars} || $self->{InterPolateEnv}) {
+    # InterPolateEnv implies InterPolateVars
+    $self->{InterPolateVars} = 1;
+
+    # we are blessing here again, to get into the ::InterPolated namespace
+    # for inheriting the methods available overthere, which we doesn't have here.
+    bless $self, 'Config::General::Interpolated';
+    eval {
+      require Config::General::Interpolated;
+    };
+    if ($EVAL_ERROR) {
+      croak "Config::General: " . $EVAL_ERROR;
+    }
+    # pre-compile the variable regexp
+    $self->{regex} = $self->_set_regex();
+  }
+#  return $self;
+}
+
+
+sub _splitpolicy {
+  #
+  # find out what split policy to use
+  my($self) = @_;
   if ($self->{SplitPolicy} ne 'guess') {
     if ($self->{SplitPolicy} eq 'whitespace') {
       $self->{SplitDelimiter} = '\s+';
@@ -233,11 +240,11 @@ sub new {
     }
     elsif ($self->{SplitPolicy} eq 'custom') {
       if (! $self->{SplitDelimiter} ) {
-	croak "SplitPolicy set to 'custom' but no SplitDelimiter set.\n";
+	croak "Config::General: SplitPolicy set to 'custom' but no SplitDelimiter set.\n";
       }
     }
     else {
-      croak "Unsupported SplitPolicy: $self->{SplitPolicy}.\n";
+      croak "Config::General: Unsupported SplitPolicy: $self->{SplitPolicy}.\n";
     }
   }
   else {
@@ -245,97 +252,124 @@ sub new {
       $self->{StoreDelimiter} = q(   );
     }
   }
-
-  if ($self->{InterPolateVars} || $self->{InterPolateEnv}) {
-    # InterPolateEnv implies InterPolateVars
-    $self->{InterPolateVars} = 1;
-
-    # we are blessing here again, to get into the ::InterPolated namespace
-    # for inheriting the methods available overthere, which we doesn't have.
-    #
-    bless $self, 'Config::General::Interpolated';
-    eval {
-      require Config::General::Interpolated;
-    };
-    if ($EVAL_ERROR) {
-      croak $EVAL_ERROR;
-    }
-    # pre-compile the variable regexp
-    $self->{regex} = $self->_set_regex();
-  }
-
-  # process as usual
-  if (!$self->{parsed}) {
-    if ($self->{DefaultConfig} && $self->{InterPolateVars}) {
-      $self->{DefaultConfig} = $self->_interpolate_hash($self->{DefaultConfig}); # FIXME: _hashref() ?
-    }
-    if (exists $self->{StringContent}) {
-      # consider the supplied string as config file
-      $self->_read($self->{StringContent}, 'SCALAR');
-      $self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
-    }
-    elsif (exists $self->{ConfigHash}) {
-      if (ref($self->{ConfigHash}) eq 'HASH') {
-	# initialize with given hash
-	$self->{config} = $self->{ConfigHash};
-	$self->{parsed} = 1;
-      }
-      else {
-	croak "Parameter -ConfigHash must be a hash reference!\n";
-      }
-    }
-    elsif (ref($self->{ConfigFile}) eq 'GLOB' || ref($self->{ConfigFile}) eq 'FileHandle') {
-      # use the file the glob points to
-      $self->_read($self->{ConfigFile});
-      $self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
-    }
-    else {
-      if ($self->{ConfigFile}) {
-	# open the file and read the contents in
-	$self->{configfile} = $self->{ConfigFile};
-	if ( file_name_is_absolute($self->{ConfigFile}) ) {
-	  # look if is is an absolute path and save the basename if it is absolute
-	  my ($volume, $path, undef) = splitpath($self->{ConfigFile});
-	  $path =~ s#/$##; # remove eventually existing trailing slash
-	  if (! $self->{ConfigPath}) {
-	    $self->{ConfigPath} = [];
-	  }
-	  unshift @{$self->{ConfigPath}}, catpath($volume, $path, q());
-	}
-	$self->_open($self->{configfile});
-	# now, we parse immdediately, getall simply returns the whole hash
-	$self->{config} = $self->_hashref();
-	$self->{config} = $self->_parse($self->{DefaultConfig}, $self->{content});
-      }
-      else {
-	# hm, no valid config file given, so try it as an empty object
-	$self->{config} = $self->_hashref();
-	$self->{parsed} = 1;
-      }
-    }
-  }
-
-  #
-  # Submodule handling. Parsing is already done at this point.
-  #
-  if ($self->{ExtendedAccess}) {
-    #
-    # we are blessing here again, to get into the ::Extended namespace
-    # for inheriting the methods available overthere, which we doesn't have.
-    #
-    bless $self, 'Config::General::Extended';
-    eval {
-      require Config::General::Extended;
-    };
-    if ($EVAL_ERROR) {
-      croak $EVAL_ERROR;
-    }
-  }
-
-  return $self;
 }
 
+sub _prepare {
+  #
+  # prepare the class parameters, mangle them, if there
+  # are options to reset or to override, do it here.
+  my ($self, %conf) = @_;
 
+  # save the parameter list for ::Extended's new() calls
+  $self->{Params} = \%conf;
+
+  # be backwards compatible
+  if (exists $conf{-file}) {
+    $self->{ConfigFile} = delete $conf{-file};
+  }
+  if (exists $conf{-hash}) {
+    $self->{ConfigHash} = delete $conf{-hash};
+  }
+
+  # store input, file, handle, or array
+  if (exists $conf{-ConfigFile}) {
+    $self->{ConfigFile} = delete $conf{-ConfigFile};
+  }
+  if (exists $conf{-ConfigHash}) {
+    $self->{ConfigHash} = delete $conf{-ConfigHash};
+  }
+
+  # store search path for relative configs, if any
+  if (exists $conf{-ConfigPath}) {
+    my $configpath = delete $conf{-ConfigPath};
+    $self->{ConfigPath} = ref $configpath eq 'ARRAY' ? $configpath : [$configpath];
+  }
+
+  # handle options which contains values we need (strings, hashrefs or the like)
+  if (exists $conf{-String} ) {
+    if (ref(\$conf{-String}) eq 'SCALAR') {
+      if ( $conf{-String}) {
+	$self->{StringContent} = $conf{-String};
+      }
+      delete $conf{-String};
+    }
+    # re-implement arrayref support, removed after 2.22 as _read were
+    # re-organized
+    elsif(ref(\$conf{-String}) eq 'ARRAY') {
+      $self->{StringContent} = join '\n', @{$conf{-String}};
+    }
+    else {
+      croak "Config::General: Parameter -String must be a SCALAR!\n";
+    }
+  }
+
+  if (exists $conf{-Tie}) {
+    if ($conf{-Tie}) {
+      $self->{Tie} = delete $conf{-Tie};
+      $self->{DefaultConfig} = $self->_hashref();
+    }
+  }
+
+  if (exists $conf{-FlagBits}) {
+    if ($conf{-FlagBits} && ref($conf{-FlagBits}) eq 'HASH') {
+      $self->{FlagBits} = 1;
+      $self->{FlagBitsFlags} = $conf{-FlagBits};
+    }
+    delete $conf{-FlagBits};
+  }
+
+  if (exists $conf{-DefaultConfig}) {
+    if ($conf{-DefaultConfig} && ref($conf{-DefaultConfig}) eq 'HASH') {
+      $self->{DefaultConfig} = $conf{-DefaultConfig};
+    }
+    elsif ($conf{-DefaultConfig} && ref($conf{-DefaultConfig}) eq q()) {
+      $self->_read($conf{-DefaultConfig}, 'SCALAR');
+      $self->{DefaultConfig} = $self->_parse($self->_hashref(), $self->{content});
+      $self->{content} = ();
+    }
+    delete $conf{-DefaultConfig};
+  }
+
+  # handle options which may either be true or false
+  # allowing "human" logic about what is true and what is not
+  foreach my $entry (keys %conf) {
+    my $key = $entry;
+    $key =~ s/^\-//;
+    if (! exists $self->{$key}) {
+      croak "Config::General: Unknown parameter: $entry => \"$conf{$entry}\" (key: <$key>)\n";
+    }
+    if ($conf{$entry} =~ /$self->{AutoTrueFlags}->{true}/io) {
+      $self->{$key} = 1;
+    }
+    elsif ($conf{$entry} =~ /$self->{AutoTrueFlags}->{false}/io) {
+      $self->{$key} = 0;
+    }
+    else {
+      # keep it untouched
+      $self->{$key} = $conf{$entry};
+    }
+  }
+
+  if ($self->{MergeDuplicateOptions}) {
+    # override if not set by user
+    if (! exists $conf{-AllowMultiOptions}) {
+      $self->{AllowMultiOptions} = 0;
+    }
+  }
+
+  if ($self->{ApacheCompatible}) {
+    # turn on all apache compatibility options which has
+    # been incorporated during the years...
+    $self->{UseApacheInclude}   = 1;
+    $self->{IncludeRelative}    = 1;
+    $self->{IncludeDirectories} = 1;
+    $self->{IncludeGlob}        = 1;
+    $self->{SpashIsDirectory}   = 1;
+    $self->{SplitPolicy}        = 'whitespace';
+    $self->{CComments}          = 0;
+    $self->{BackslashEscape}    = 1;
+  }
+}
 
 sub getall {
   #
@@ -359,8 +393,16 @@ sub _open {
   #
   # open the config file, or expand a directory or glob
   #
-  my($this, $configfile) = @_;
-  my $fh;
+  my($this, $basefile, $basepath) = @_;
+  my($fh, $configfile);
+
+  if($basepath) {
+    # if this doesn't work we can still try later the global config path to use
+    $configfile = catfile($basepath, $basefile);
+  }
+  else {
+    $configfile = $basefile;
+  }
 
   if ($this->{IncludeGlob} and $configfile =~ /[*?\[\{\\]/) {
     # Something like: *.conf (or maybe dir/*.conf) was included; expand it and
@@ -385,8 +427,8 @@ sub _open {
     if (defined $this->{ConfigPath}) {
       # try to find the file within ConfigPath
       foreach my $dir (@{$this->{ConfigPath}}) {
-	if( -e catfile($dir, $configfile) ) {
-	  $configfile = catfile($dir, $configfile);
+	if( -e catfile($dir, $basefile) ) {
+	  $configfile = catfile($dir, $basefile);
 	  $found = 1;
 	  last; # found it
 	}
@@ -394,7 +436,7 @@ sub _open {
     }
     if (!$found) {
       my $path_message = defined $this->{ConfigPath} ? q( within ConfigPath: ) . join(q(.), @{$this->{ConfigPath}}) : q();
-      croak qq{The file "$configfile" does not exist$path_message!};
+      croak qq{Config::GeneralThe file "$basefile" does not exist$path_message!};
     }
   }
 
@@ -407,26 +449,31 @@ sub _open {
   if (-d $configfile and $this->{IncludeDirectories}) {
     # A directory was included; include all the files inside that directory in ASCII order
     local *INCLUDEDIR;
-    opendir INCLUDEDIR, $configfile or croak "Could not open directory $configfile!($!)\n";
-    my @files = sort grep { -f "$configfile/$_" } "$configfile/$_", readdir INCLUDEDIR;
+    opendir INCLUDEDIR, $configfile or croak "Config::General: Could not open directory $configfile!($!)\n";
+    my @files = sort grep { -f catfile($configfile, $_) } catfile($configfile, $_), readdir INCLUDEDIR;
     closedir INCLUDEDIR;
     local $this->{CurrentConfigFilePath} = $configfile;
     for (@files) {
-      if (! $this->{files}->{"$configfile/$_"}) {
-	$fh = IO::File->new( "$configfile/$_", 'r') or croak "Could not open $configfile/$_!($!)\n";
-	$this->{files}->{"$configfile/$_"} = 1;
+      my $file = catfile($configfile, $_);
+      if (! exists $this->{files}->{$file} or $this->{IncludeAgain} ) {
+        # support re-read if used urged us to do so, otherwise ignore the file
+	$fh = IO::File->new( $file, 'r') or croak "Config::General: Could not open $file!($!)\n";
+	$this->{files}->{"$file"} = 1;
 	$this->_read($fh);
       }
+      else {
+        warn "File $file already loaded.  Use -IncludeAgain to load it again.\n";
+      }
     }
   }
   elsif (-e _) {
-    if (exists $this->{files}->{$configfile} ) {
+    if (exists $this->{files}->{$configfile} and not $this->{IncludeAgain}) {
       # do not read the same file twice, just return
-      # FIXME: should we croak here, when some "debugging" is enabled?
+      warn "File $configfile already loaded.  Use -IncludeAgain to load it again.\n";
       return;
     }
     else {
-      $fh = IO::File->new( "$configfile", 'r') or croak "Could not open $configfile!($!)\n";
+      $fh = IO::File->new( "$configfile", 'r') or croak "Config::General: Could not open $configfile!($!)\n";
 
       $this->{files}->{$configfile}    = 1;
 
@@ -555,7 +602,7 @@ sub _read {
 
 
     # transform explicit-empty blocks to conforming blocks
-    if (/\s*<([^\/]+?.*?)\/>$/) {
+    if (!$this->{ApacheCompatible} && /\s*<([^\/]+?.*?)\/>$/) {
       my $block = $1;
       if ($block !~ /\"/) {
 	if ($block !~ /\s[^\s]/) {
@@ -623,7 +670,7 @@ sub _read {
 	$incl_file = $1;
 	if ( $this->{IncludeRelative} && $path && !file_name_is_absolute($incl_file) ) {
 	  # include the file from within location of $this->{configfile}
-	  $this->_open( catfile($path, $incl_file) );
+	  $this->_open( $incl_file, $path );
 	}
 	else {
 	  # include the file from within pwd, or absolute
@@ -700,9 +747,9 @@ sub _parse {
 	if ($this->{InterPolateVars}) {
 	  # interpolate block(name), add "<" and ">" to the key, because
 	  # it is sure that such keys does not exist otherwise.
-	  $block     = $this->_interpolate("<$block>", $block);
+	  $block     = $this->_interpolate($config, "<$block>", $block);
 	  if (defined $blockname) {
-	    $blockname = $this->_interpolate("<$blockname>", "$blockname");
+	    $blockname = $this->_interpolate($config, "<$blockname>", "$blockname");
 	  }
 	}
 	if ($this->{LowerCaseNames}) {
@@ -712,21 +759,22 @@ sub _parse {
 	undef @newcontent;
 	next;
       }
-      elsif (/^<\/(.+?)>$/) {                              # it is an end block, but we don't have a matching block!
-	croak "EndBlock \"<\/$1>\" has no StartBlock statement (level: $this->{level}, chunk $chunk)!\n";
+      elsif (/^<\/(.+?)>$/) { # it is an end block, but we don't have a matching block!
+	croak "Config::General: EndBlock \"<\/$1>\" has no StartBlock statement (level: $this->{level}, chunk $chunk)!\n";
       }
       else {                                               # insert key/value pair into actual node
 	if ($this->{LowerCaseNames}) {
 	  $option = lc $option;
 	}
+
 	if (exists $config->{$option}) {
 	  if ($this->{MergeDuplicateOptions}) {
-	    $config->{$option} = $this->_parse_value($option, $value);
+	    $config->{$option} = $this->_parse_value($config, $option, $value);
 	  }
 	  else {
 	    if (! $this->{AllowMultiOptions} ) {
 	      # no, duplicates not allowed
-	      croak "Option \"$option\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
+	      croak "Config::General: Option \"$option\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
 	    }
 	    else {
 	      # yes, duplicates allowed
@@ -740,18 +788,23 @@ sub _parse {
 		my $i = scalar @{$config->{$option}};
 	      };
 	      if ($EVAL_ERROR) {
-		$config->{$option} = $this->_parse_value($option, $value);
+		$config->{$option} = $this->_parse_value($config, $option, $value);
 	      }
 	      else {
 		# it's already an array, just push
-		push @{$config->{$option}}, $this->_parse_value($option, $value);
+		push @{$config->{$option}}, $this->_parse_value($config, $option, $value);
 	      }
 	    }
 	  }
 	}
 	else {
 	  # standard config option, insert key/value pair into node
-	  $config->{$option} = $this->_parse_value($option, $value);
+	  $config->{$option} = $this->_parse_value($config, $option, $value);
+
+	  if ($this->{InterPolateVars}) {
+	    # save pair on local stack
+	    $config->{__stack}->{$option} = $config->{$option};
+	  }
 	}
       }
     }
@@ -765,14 +818,29 @@ sub _parse {
 	push @newcontent, $_;       # push onto new content stack
       }
       else {                        # calling myself recursively, end of $block reached, $block_level is 0
-	if (defined $blockname) {           # a named block, make it a hashref inside a hash within the current node
-	  $this->_savelast($blockname);
+	if (defined $blockname) {
+	  # a named block, make it a hashref inside a hash within the current node
 
 	  if (! exists $config->{$block}) {
-	    $config->{$block} = $this->_hashref(); # Make sure that the hash is not created implicitely
+	    # Make sure that the hash is not created implicitly
+	    $config->{$block} = $this->_hashref();
+
+	    if ($this->{InterPolateVars}) {
+	      # inherit current __stack to new block
+	      $config->{$block}->{__stack} = $config->{__stack};
+	    }
 	  }
 
-	  if (exists $config->{$block}->{$blockname}) {    # the named block already exists, make it an array
+	  if (ref($config->{$block}) eq '') {
+	    croak "Config::General: Block <$block> already exists as scalar entry!\n";
+	  }
+	  elsif (ref($config->{$block}) eq 'ARRAY') {
+	    croak "Config::General: Cannot append named block <$block $blockname> to array of scalars!\n"
+	         ."Block <$block> or scalar '$block' occurs more than once.\n"
+	         ."Turn on -MergeDuplicateBlocks or make sure <$block> occurs only once in the config.\n";
+	  }
+	  elsif (exists $config->{$block}->{$blockname}) {
+	    # the named block already exists, make it an array
 	    if ($this->{MergeDuplicateBlocks}) {
 	      # just merge the new block with the same name as an existing one into
               # this one.
@@ -780,7 +848,7 @@ sub _parse {
 	    }
 	    else {
 	      if (! $this->{AllowMultiOptions}) {
-		croak "Named block \"<$block $blockname>\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
+		croak "Config::General: Named block \"<$block $blockname>\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
 	      }
 	      else {                                       # preserve existing data
 		my $savevalue = $config->{$block}->{$blockname};
@@ -797,19 +865,27 @@ sub _parse {
 	      }
 	    }
 	  }
-	  elsif (ref($config->{$block}) eq 'ARRAY') {
-	    croak "Cannot add named block <$block $blockname> to hash! Block <$block> occurs more than once.\n"
-	         ."Turn on -MergeDuplicateBlocks or make sure <$block> occurs only once in the config.\n";
-	  }
 	  else {
 	    # the first occurence of this particular named block
-	    $config->{$block}->{$blockname} = $this->_parse($this->_hashref(), \@newcontent);
+	    my $tmphash = $this->_hashref();
+
+	    if ($this->{InterPolateVars}) {
+	      # inherit current __stack to new block
+	      $tmphash->{__stack} = $config->{__stack};
+	    }
+
+	    $config->{$block}->{$blockname} = $this->_parse($tmphash, \@newcontent);
 	  }
-	  $this->_backlast($blockname);
 	}
-	else {                                             # standard block
-	  $this->_savelast($block);
-	  if (exists $config->{$block}) {                  # the block already exists, make it an array
+	else {
+	  # standard block
+	  if (exists $config->{$block}) {
+	    if (ref($config->{$block}) eq '') {
+	      croak "Config::General: Cannot create hashref from <$block> because there is\n"
+		   ."already a scalar option '$block' with value '$config->{$block}'\n";
+	    }
+
+	    # the block already exists, make it an array
 	    if ($this->{MergeDuplicateBlocks}) {
 	      # just merge the new block with the same name as an existing one into
               # this one.
@@ -817,7 +893,7 @@ sub _parse {
             }
             else {
 	      if (! $this->{AllowMultiOptions}) {
-	        croak "Block \"<$block>\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
+	        croak "Config::General: Block \"<$block>\" occurs more than once (level: $this->{level}, chunk $chunk)!\n";
 	      }
 	      else {
 		my $savevalue = $config->{$block};
@@ -836,10 +912,15 @@ sub _parse {
 	  }
 	  else {
 	    # the first occurence of this particular block
-	    #### $config->{$block} = $this->_parse($config->{$block}, \@newcontent);
-	    $config->{$block} = $this->_parse($this->_hashref(), \@newcontent);
+	    my $tmphash = $this->_hashref();
+
+	    if ($this->{InterPolateVars}) {
+	      # inherit current __stack to new block
+	      $tmphash->{__stack} = $config->{__stack};
+	    }
+
+	    $config->{$block} = $this->_parse($tmphash, \@newcontent);
 	  }
-	  $this->_backlast($block);
 	}
 	undef $blockname;
 	undef $block;
@@ -854,34 +935,19 @@ sub _parse {
   if ($block) {
     # $block is still defined, which means, that it had
     # no matching endblock!
-    croak "Block \"<$block>\" has no EndBlock statement (level: $this->{level}, chunk $chunk)!\n";
+    croak "Config::General: Block \"<$block>\" has no EndBlock statement (level: $this->{level}, chunk $chunk)!\n";
   }
   return $config;
 }
 
 
-sub _savelast {
-  my($this, $key) = @_;
-  push @{$this->{upperkeys}}, $this->{lastkey};
-  $this->{lastkey}  = $this->{prevkey};
-  $this->{prevkey}  = $key;
-  return;
-}
-
-sub _backlast {
-  my($this, $key) = @_;
-  $this->{prevkey} = $this->{lastkey};
-  $this->{lastkey} = pop @{$this->{upperkeys}};
-  return;
-}
-
 sub _parse_value {
   #
   # parse the value if value parsing is turned on
   # by either -AutoTrue and/or -FlagBits
   # otherwise just return the given value unchanged
   #
-  my($this, $option, $value) =@_;
+  my($this, $config, $option, $value) =@_;
 
   # avoid "Use of uninitialized value"
   if (! defined $value) {
@@ -889,7 +955,7 @@ sub _parse_value {
   }
 
   if ($this->{InterPolateVars}) {
-    $value = $this->_interpolate($option, $value);
+    $value = $this->_interpolate($config, $option, $value);
   }
 
   # make true/false values to 1 or 0 (-AutoTrue)
@@ -931,7 +997,7 @@ sub NoMultiOptions {
   # Since we do parsing from within new(), we must
   # call it again if one turns NoMultiOptions on!
   #
-  croak q(The NoMultiOptions() method is deprecated. Set 'AllowMultiOptions' to 'no' instead!);
+  croak q(Config::GeneralThe NoMultiOptions() method is deprecated. Set 'AllowMultiOptions' to 'no' instead!);
 }
 
 
@@ -951,7 +1017,7 @@ sub save {
     $this->save_file($one, \%h);
   }
   else {
-    croak q(The save() method is deprecated. Use the new save_file() method instead!);
+    croak q(Config::GeneralThe save() method is deprecated. Use the new save_file() method instead!);
   }
   return;
 }
@@ -966,17 +1032,17 @@ sub save_file {
   my $config_string;
 
   if (!$file) {
-    croak "Filename is required!";
+    croak "Config::General: Filename is required!";
   }
   else {
-    $fh = IO::File->new( "$file", 'w') or croak "Could not open $file!($!)\n";
+    $fh = IO::File->new( "$file", 'w') or croak "Config::General: Could not open $file!($!)\n";
 
     if (!$config) {
       if (exists $this->{config}) {
 	$config_string = $this->_store(0, %{$this->{config}});
       }
       else {
-	croak "No config hash supplied which could be saved to disk!\n";
+	croak "Config::General: No config hash supplied which could be saved to disk!\n";
       }
     }
     else {
@@ -1009,7 +1075,7 @@ sub save_string {
       return $this->_store(0, %{$this->{config}});
     }
     else {
-      croak "No config hash supplied which could be saved to disk!\n";
+      croak "Config::General: No config hash supplied which could be saved to disk!\n";
     }
   }
   else {
@@ -1030,7 +1096,7 @@ sub _store {
 
   my $config_string = q();
 
-  foreach my $entry (sort keys %config) {
+  foreach my $entry (keys %config) {
     if (ref($config{$entry}) eq 'ARRAY') {
       foreach my $line (@{$config{$entry}}) {
         if (ref($line) eq 'HASH') {
@@ -1125,10 +1191,10 @@ sub _hashref {
       $wantarray, $evaltext, $is_require, $hints, $bitmask) = caller 0;
   if ($this->{Tie}) {
     eval {
-      eval {require $this->{Tie}};
+      eval qq{require $this->{Tie}};
     };
     if ($EVAL_ERROR) {
-      croak q(Could not create a tied hash of type: ) . $this->{Tie} . q(: ) . $EVAL_ERROR;
+      croak q(Config::GeneralCould not create a tied hash of type: ) . $this->{Tie} . q(: ) . $EVAL_ERROR;
     }
     my %hash;
     tie %hash, $this->{Tie};
@@ -1158,11 +1224,11 @@ sub SaveConfig {
   my ($file, $hash) = @_;
 
   if (!$file || !$hash) {
-    croak q{SaveConfig(): filename and hash argument required.};
+    croak q{Config::GeneralSaveConfig(): filename and hash argument required.};
   }
   else {
     if (ref($hash) ne 'HASH') {
-      croak q(The second parameter must be a reference to a hash!);
+      croak q(Config::GeneralThe second parameter must be a reference to a hash!);
     }
     else {
       (new Config::General(-ConfigHash => $hash))->save_file($file);
@@ -1179,11 +1245,11 @@ sub SaveConfigString {
   my ($hash) = @_;
 
   if (!$hash) {
-    croak q{SaveConfigString(): Hash argument required.};
+    croak q{Config::GeneralSaveConfigString(): Hash argument required.};
   }
   else {
     if (ref($hash) ne 'HASH') {
-      croak q(The parameter must be a reference to a hash!);
+      croak q(Config::GeneralThe parameter must be a reference to a hash!);
     }
     else {
       return (new Config::General(-ConfigHash => $hash))->save_string();
@@ -1350,6 +1416,19 @@ with standard file patterns, * will not match dot-files, so <<include dir/*>>
 is often more desirable than including a directory with B<-IncludeDirectories>.
 
 
+=item B<-IncludeAgain>
+
+If set to a true value, you will be able to include a sub-configfile
+multiple times.  With the default, false, you will get a warning about
+duplicate includes and only the first include will succeed.
+
+Reincluding a configfile can be useful if it contains data that you want to
+be present in multiple places in the data tree.  See the example under
+L</INCLUDES>.
+
+Note, however, that there is currently no check for include recursion.
+
+
 =item B<-ConfigPath>
 
 As mentioned above, you can use this variable to specify a search path for relative
@@ -1652,6 +1731,35 @@ Please note that this is a new option (incorporated in version 2.30),
 it may lead to various unexpected sideeffects or other failures.
 You've been warned.
 
+=item B<-ApacheCompatible>
+
+Over the past years a lot of options has been incorporated
+into Config::General to be able to parse real apache configs.
+
+The new B<-ApacheCompatible> option now makes it possible to
+tweak all options in a way that apache configs can be parsed.
+
+This is called "apache compatibility mode" - if you will ever
+have problems with parsing apache configs without this option
+being set, you'll get no help by me. Thanks :)
+
+The following options will be set:
+
+ UseApacheInclude   = 1
+ IncludeRelative    = 1
+ IncludeDirectories = 1
+ IncludeGlob        = 1
+ SpashIsDirectory   = 1
+ SplitPolicy        = 'equalsign'
+ CComments          = 0
+ BackslashEscape    = 1
+
+Take a look into the particular documentation sections what
+those options are doing.
+
+Beside setting some options it also turns off support for
+explicit empty blocks.
+
 =back
 
 
@@ -2107,6 +2215,34 @@ Include statements can be case insensitive (added in version 1.25).
 
 Include statements will be ignored within C-Comments and here-documents.
 
+By default, a config file will only be included the first time it is
+referenced.  If you wish to include a file in multiple places, set
+B</-IncludeAgain> to true. But be warned: this may lead to infinite loops,
+so make sure, you're not including the same file from within itself!
+
+Example:
+
+    # main.cfg
+    <object billy>
+        class=Some::Class
+        <printers>
+            include printers.cfg
+        </printers>
+        # ...
+    </object>
+    <object bob>
+        class=Another::Class
+        <printers>
+            include printers.cfg
+        </printers>
+        # ...
+    </object>
+
+Now C<printers.cfg> will be include in both the C<billy> and C<bob> objects.
+
+You will have to be careful to not recursively include a file.  Behaviour
+in this case is undefined.
+
 
 
 =head1 COMMENTS
@@ -2175,7 +2311,7 @@ allowed to the B<new()> method of the standard interface.
 
 Example:
 
- use Config::General;
+ use Config::General qw(ParseConfig);
  my %config = ParseConfig(-ConfigFile => "rcfile", -AutoTrue => 1);
 
 
@@ -2186,7 +2322,7 @@ to a hash structure.
 
 Example:
 
- use Config::General;
+ use Config::General qw(SaveConfig);
  ..
  SaveConfig("rcfile", \%some_hash);
 
@@ -2199,7 +2335,7 @@ method B<save_string()> does.
 
 Example:
 
- use Config::General;
+ use Config::General qw(ParseConfig SaveConfigString);
  my %config = ParseConfig(-ConfigFile => "rcfile");
  .. # change %config something
  my $content = SaveConfigString(\%config);
@@ -2254,7 +2390,7 @@ Thomas Linden <tlinden |AT| cpan.org>
 
 =head1 VERSION
 
-2.32
+2.37
 
 =cut
 
